"""Provides scales"""

from plotnine.scales.scale import scale_discrete
from .pal import prism_color_pal, prism_fill_pal, prism_shape_pal


class scale_color_prism(scale_discrete):
    """Prism color scale

    Args:
        palette: The color palette name
    """

    _aesthetics = ["color"]
    na_value = "#7F7F7F"

    def __init__(self, palette="colors", **kwargs):
        """Construct"""
        self.palette = prism_color_pal(palette)
        scale_discrete.__init__(self, **kwargs)


class scale_fill_prism(scale_color_prism):
    """Prism fill scale

    Args:
        palette: The fill palette name
    """

    _aesthetics = ["fill"]
    na_value = "#7F7F7F"

    def __init__(self, palette="colors", **kwargs):
        """Construct"""
        self.palette = prism_fill_pal(palette)
        scale_discrete.__init__(self, **kwargs)


class scale_shape_prism(scale_discrete):
    """Prism shape scale

    Args:
        palette: The shape palette name
    """

    _aesthetics = ["shape"]

    def __init__(self, palette="default", **kwargs):
        """Construct"""
        self.palette = prism_shape_pal(palette)
        scale_discrete.__init__(self, **kwargs)


scale_colour_prism = scale_color_prism



"""Provides theme_prism"""
from plotnine.options import get_option
from plotnine.themes.elements import (
    element_blank,
    element_line,
    element_rect,
    element_text,
)
from plotnine.themes.theme import theme as theme_abc

from .pal import list_themes, theme_colors

# Make sure the themeables work
from .themeable import (  # noqa: F401
    axis_ticks_x,
    axis_ticks_y,
    prism_ticks_length,
    prism_ticks_length_x,
    prism_ticks_length_y,
)


class theme_prism(theme_abc):
    """The theme_prism() class

    Args:
        palette: The theme palette
        base_size: The base size of the plot
        base_family: The base font family
        base_fontface: The base font face
        base_line_size: The base line size
        base_rect_size: The base rect size
        axis_text_angle: The default axis text angle
        border: Whether to show border of the plot
        **kwargs: Other theming options
    """

    def __init__(
        self,
        palette="black_and_white",
        base_size=10.0,
        base_family="DejaVu Sans",
        base_fontface="bold",
        base_line_size=None,
        base_rect_size=None,
        axis_text_angle=0,
        border=False,
        **kwargs,
    ):
        """Construct"""
        if axis_text_angle not in (0, 45, 90, 270):
            raise ValueError(
                "'axis_text_angle' must be one of [0, 45, 90, 270].\n"
                "For other angles, use `guide_axis()` function from plotnine "
                "instead"
            )

        if palette not in list_themes():
            raise ValueError(
                f"The palette '{palette}' does not exist. "
                f"Supported palettes are: {list_themes()}"
            )

        if base_line_size is None:
            base_line_size = 4 / 3.0 * base_size / 10.0
        if base_rect_size is None:
            base_rect_size = 4 / 3.0 * base_size / 10.0

        colours = theme_colors(palette)
        if border:
            panel_border = element_rect(fill=None)
            axis_line = element_blank()
        else:
            panel_border = element_blank()
            axis_line = element_line()

        theme_abc.__init__(
            self,
            # Base elements (to be inherited by other elements)
            line=element_line(
                colour=colours.axisColor,
                size=base_line_size,
                linetype="solid",
                lineend="butt",
            ),
            rect=element_rect(
                fill="white",
                colour=colours.axisColor,
                size=base_rect_size,
                linetype="solid",
            ),
            text=element_text(
                family=base_family,
                face=base_fontface,
                colour=colours.graphTitleColor,
                size=base_size,
                lineheight=0.9,
                hjust=0.5,
                vjust=0.5,
                angle=0,
                margin={},
            ),
            # Options
            dpi=get_option("dpi"),
            figure_size=get_option("figure_size"),
            # Prism custom theme elements
            prism_ticks_length=base_size / 5,
            # Normal ggplot2 theme elements
            axis_line=axis_line,
            axis_line_x=None,
            axis_line_y=None,
            axis_text=element_text(
                size=0.95 * base_size, colour=colours.axisLabelColor
            ),
            axis_text_x=element_text(
                margin=dict(t=0.8 * base_size),
                angle=axis_text_angle,
                hjust=2 if axis_text_angle in (45, 90, 270) else 0.5,
                vjust=0.5 if axis_text_angle in (0, 90, 270) else 1,
            ),
            # axis_text_x_top=element_text(margin=dict(b=0.8 * base_size / 4),
            # vjust=0),
            axis_text_y=element_text(
                margin=dict(r=0.8 * base_size / 4), hjust=1
            ),
            # axis_text_y_right=element_text(
            # margin=dict(l=0.5 * base_size / 4),
            # hjust=0),
            axis_ticks=element_line(),
            # axis_ticks_length =  unit(base_size / 2.5, "pt"),
            axis_ticks_length=base_size / 2.5,
            # axis_ticks_length_x=None,
            # axis_ticks_length_x_top=None,
            # axis_ticks_length_x_bottom=None,
            # axis_ticks_length_y=None,
            # axis_ticks_length_y_top=None,
            # axis_ticks_length_y_bottom=None,
            axis_title=element_text(colour=colours.axisTitleColor),
            axis_title_x=element_text(margin=dict(t=base_size * 0.6), vjust=1),
            # axis_title_x_top=element_text(margin=dict(b=base_size * 0.6),
            # vjust=0),
            axis_title_y=element_text(
                angle=90, margin=dict(r=base_size * 1.5), vjust=1
            ),
            # axis_title_y_right=element_text(
            #     angle=-90, margin=dict(l=base_size * 0.6), vjust=0
            # ),
            # legend
            legend_background=element_rect(color="None", fill="None"),
            legend_entry_spacing_x=None,
            legend_entry_spacing_y=None,
            legend_key=element_rect(colour="None", fill="None"),
            legend_key_size=base_size * 0.8 * 1.8,
            legend_key_height=None,
            legend_key_width=base_size * 1.5,
            legend_margin=0,  # points
            legend_spacing=base_size,  # points
            legend_text=element_text(
                size=base_size * 0.75,
                ha="left",
                weight="normal",
                margin={"t": 3, "b": 3, "l": 3, "r": 3, "units": "pt"},
            ),
            legend_text_legend=element_text(va="baseline", weight="normal"),
            legend_text_colorbar=element_text(va="center", weight="normal"),
            legend_title=element_blank(),
            legend_title_align=None,
            legend_position="right",
            legend_box=None,
            legend_box_margin=0,  # points
            legend_box_just=None,
            legend_box_spacing=0.1,  # In inches
            legend_direction=None,
            panel_background=element_rect(
                fill=colours.plottingAreaColor
                if palette == "office"
                else "None",
                colour=None,
            ),
            panel_border=panel_border,
            panel_grid=element_blank(),
            panel_grid_minor=element_blank(),
            # panel_spacing =      unit(base_size / 2, "pt"),
            panel_spacing=4 * base_size / 3.0,
            panel_spacing_x=None,
            panel_spacing_y=None,
            panel_ontop=False,
            strip_background=element_blank(),
            strip_text=element_text(
                colour=colours.axisTitleColor,
                # size = rel(0.8),
                size=0.8 * base_size,
                margin=dict(
                    t=base_size / 2.5,
                    b=base_size / 2.5,
                    l=base_size / 2.5,
                    r=base_size / 2.5,
                ),
            ),
            strip_text_x=element_text(margin=dict(b=base_size / 3)),
            strip_text_y=element_text(angle=-90, margin=dict(l=base_size / 3)),
            # strip_text_y_left=element_text(angle=90),
            # strip_placement="inside",
            # strip_placement_x=None,
            # strip_placement_y=None,
            # strip_switch_pad_grid = unit(base_size / 4, "pt"),
            # strip_switch_pad_wrap = unit(base_size / 4, "pt"),
            # strip_switch_pad_grid=base_size / 4,
            # strip_switch_pad_wrap=base_size / 4,
            plot_background=element_rect(
                fill=colours.pageBackgroundColor,
                colour=colours.pageBackgroundColor,
            ),
            # plot_title =         element_text(size = rel(1.2),
            plot_title=element_text(
                size=1.2 * base_size,
                hjust=0.5,
                vjust=1,
                margin=dict(b=base_size),
            ),
            # plot_title_position="panel",
            # plot_subtitle=element_text(
            #     hjust=0.5, vjust=1, margin=dict(b=base_size / 2)
            # ),
            # plot_caption =       element_text(size = rel(0.8),
            # plot_caption=element_text(
            #     size=0.8, hjust=1, vjust=1, margin=dict(t=base_size / 2)
            # ),
            # plot_caption_position="panel",
            # plot_tag =           element_text(size = rel(1.2),
            # plot_tag=element_text(size=1.2, hjust=0.5, vjust=0.5),
            # plot_tag_position="topleft",
            # plot_margin=base_size / 2,
            plot_margin=None,
            complete=True,
            **kwargs,
        )



"""Prism themes for plotnine, inspired by ggprism"""

from .theme import theme_prism
from .scale import (
    scale_color_prism,
    scale_fill_prism,
    scale_shape_prism,
    scale_colour_prism,
)
from .pal import (
    list_themes,
    list_color_pals,
    list_colour_pals,
    list_fill_pals,
    list_shape_pals,
    prism_color_pal,
    prism_colour_pal,
    prism_fill_pal,
    prism_shape_pal,
)
from .guide import (
    guide_prism,
    guide_prism_minor,
    # guide_prism_bracket,
    guide_prism_offset,
    guide_prism_offset_minor,
)

__all__ = (
    "theme_prism",
    "scale_color_prism",
    "scale_colour_prism",
    "scale_fill_prism",
    "scale_shape_prism",
    "list_themes",
    "list_color_pals",
    "list_colour_pals",
    "list_fill_pals",
    "list_shape_pals",
    "prism_color_pal",
    "prism_colour_pal",
    "prism_fill_pal",
    "prism_shape_pal",
    "guide_prism",
    "guide_prism_minor",
    # "guide_prism_bracket",
    "guide_prism_offset",
    "guide_prism_offset_minor",
)

__version__ = "0.1.0"



"""Add themeables for theme_prism"""

from contextlib import suppress
from copy import deepcopy
from plotnine.themes.themeable import (
    themeable as themeable_abc,
    axis_ticks_major_x,
    axis_ticks_major_y,
    axis_ticks_minor_x,
    axis_ticks_minor_y,
)


class axis_ticks_x(axis_ticks_major_x, axis_ticks_minor_x):
    """Themeable for ticks on x axis that is missing in plotnine"""

    def apply(self, ax):
        """Apply themeable to the axis"""
        axis_ticks_major_x.apply(self, ax)
        axis_ticks_minor_x.apply(self, ax)

    def blank(self, ax):
        """When it's set blank"""
        axis_ticks_major_x.blank(self, ax)
        axis_ticks_minor_x.blank(self, ax)


class axis_ticks_y(axis_ticks_major_y, axis_ticks_minor_y):
    """Themeable for ticks on y axis that is missing in plotnine"""

    def apply(self, ax):
        """Apply themeable to the axis"""
        axis_ticks_major_y.apply(self, ax)
        axis_ticks_minor_y.apply(self, ax)

    def blank(self, ax):
        """When it's set blank"""
        axis_ticks_major_y.blank(self, ax)
        axis_ticks_minor_y.blank(self, ax)


class prism_ticks_length_x(themeable_abc):
    """Themeable for ticks length on x axis"""

    def apply(self, ax):
        """Apply themeable to the axis"""
        themeable_abc.apply(self, ax)

        d = deepcopy(self.properties)
        with suppress(KeyError):
            length = d.pop("value")
            ax.xaxis.set_tick_params(
                which="minor",
                length=abs(length),
                direction="in" if length < 0 else "out",
            )

        for tick in ax.xaxis.get_minor_ticks():
            tick.tick1line.set(**d)

    def blank(self, ax):
        """When it's set blank"""
        themeable_abc.blank(self, ax)
        ax.xaxis.set_tick_params(which="minor", bottom=False)


class prism_ticks_length_y(themeable_abc):
    """Themeable for ticks length on y axis"""

    def apply(self, ax):
        """Apply themeable to the axis"""
        themeable_abc.apply(self, ax)

        d = deepcopy(self.properties)
        with suppress(KeyError):
            length = d.pop("value")
            ax.yaxis.set_tick_params(
                which="minor",
                length=abs(length),
                direction="in" if length < 0 else "out",
            )

        for tick in ax.yaxis.get_minor_ticks():
            tick.tick1line.set(**d)

    def blank(self, ax):
        """When it's set blank"""
        themeable_abc.blank(self, ax)
        ax.yaxis.set_tick_params(which="minor", bottom=False)


class prism_ticks_length(prism_ticks_length_x, prism_ticks_length_y):
    """Themeable for ticks length"""

    def apply(self, ax):
        """Apply themeable to the axis"""
        prism_ticks_length_x.apply(self, ax)
        prism_ticks_length_y.apply(self, ax)

    def blank(self, ax):
        """When it's set blank"""
        prism_ticks_length_x.blank(self, ax)
        prism_ticks_length_y.blank(self, ax)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['plotnine_prism']

package_data = \
{'': ['*'], 'plotnine_prism': ['schemes/*']}

install_requires = \
['diot>=0.1,<0.2', 'plotnine>=0.8,<0.9', 'rtoml>=0.8,<0.9']

setup_kwargs = {
    'name': 'plotnine-prism',
    'version': '0.1.0',
    'description': 'Prism themes for plotnine, inspired by ggprism',
    'long_description': '# plotnine-prism\n\nPrism themes for [plotnine][1], inspired by [ggprism][2].\n\n\n## Installation\n\n```\npip install -U plotnine-prism\n```\n\n## Documentation\n\n[https://pwwang.github.io/plotnine-prism][3]\n\n## Usage\n\nSee [this notebook][6] for the following example, and also [Getting started][4] for a quick overview of `plotnine_prism` features.\n\n```python\nfrom plotnine import *\nfrom plotnine_prism import *\n\nfrom datar.all import f, as_categorical, mutate\nfrom datar.datasets import ToothGrowth\n\nToothGrowth >>= mutate(dose=as_categorical(f.dose))\n\nbase = (\n    ggplot(ToothGrowth, aes(x = "dose", y = "len")) +\n    geom_violin(aes(colour = "dose", fill = "dose"), trim = False) +\n    geom_boxplot(aes(fill = "dose"), width = 0.2, colour = "black")\n)\n\np1 = base + ylim(-5, 40)\np2 = (\n    base +\n    scale_y_continuous(limits=[-5, 40], guide=guide_prism_offset_minor()) +\n    scale_color_prism(\'floral\') +\n    scale_fill_prism(\'floral\') +\n    theme_prism()\n)\n# See examples/nb_helpers.py for plot_grid function\n# plot_grid(p1, p2)\n```\n\n<table>\n    <tr>\n        <td valign="top"><img src="./toothgrowth1.png" /></td>\n        <td valign="top"><img src="./toothgrowth2.png" /></td>\n    </tr>\n</table>\n\n## More examples\n\nThe Dose Response Curve was recreated. See [this vignette][5] for the source code and step-by-step instructions.\n\n\n<table>\n    <tr>\n        <td valign="top"><img src="./dose1.png" /></td>\n        <td valign="top"><img src="./dose2.png" /></td>\n    </tr>\n</table>\n\n[1]: https://github.com/has2k1/plotnine\n[2]: https://github.com/csdaw/ggprism/\n[3]: https://pwwang.github.io/plotnine-prism\n[4]: https://pwwang.github.io/plotnine-prism/get_started\n[5]: https://pwwang.github.io/plotnine-prism/raw/ex1-dose\n[6]: https://pwwang.github.io/plotnine-prism/raw/README\n\n',
    'author': 'pwwang',
    'author_email': 'pwwang@pwwang.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Provides guides"""
import hashlib
from warnings import warn

from matplotlib.ticker import NullLocator

from plotnine.guides.guide import guide
from plotnine.exceptions import PlotnineWarning


class guide_prism_offset(guide):
    """The prism offset guide that offset the axes"""

    offset = 20
    aesthetic = None
    hash = None
    available_aes = {"x", "y"}

    def train(self, scale, aesthetic=None):
        """Get arguments from the scale"""
        if aesthetic is None:
            aesthetic = scale.aesthetics[0]

        # Do nothing if scales are inappropriate
        if set(scale.aesthetics) & self.available_aes == 0:
            warn(
                f"'{self.__class__.__name__}' needs appropriate scales.",
                PlotnineWarning,
            )
            return None

        self.aesthetic = list(set(scale.aesthetics) & self.available_aes)[0]

        info = "\n".join([self.title, self.__class__.__name__])
        self.hash = hashlib.md5(info.encode("utf-8")).hexdigest()
        return self

    def create_geoms(self, plot):
        """Apply the arguments"""
        xaxis_position = plot.axs[0].xaxis.get_label_position()
        yaxis_position = plot.axs[0].yaxis.get_label_position()
        gca = plot.axs[0].axes

        if self.aesthetic == "x":
            major_locs = gca.get_xticks()
            minor_locs = gca.xaxis.get_minorticklocs()
            if yaxis_position == "left":
                offset = (
                    self.offset
                    if major_locs[0] > minor_locs[0]
                    else self.offset / 2
                )
            else:
                offset = (
                    self.offset
                    if major_locs[-1] < minor_locs[-1]
                    else self.offset / 2
                )
            gca.spines[yaxis_position].set_position(("outward", offset))
            gca.set_xlim(min(major_locs), max(major_locs))
            gca.xaxis.set_minor_locator(NullLocator())
        else:
            major_locs = gca.get_yticks()
            minor_locs = gca.yaxis.get_minorticklocs()
            if xaxis_position == "bottom":
                offset = (
                    self.offset
                    if major_locs[0] > minor_locs[0]
                    else self.offset / 2
                )
            else:
                offset = (
                    self.offset
                    if major_locs[-1] < minor_locs[-1]
                    else self.offset / 2
                )
            gca.spines[xaxis_position].set_position(("outward", offset))
            gca.set_ylim(min(major_locs), max(major_locs))
            gca.yaxis.set_minor_locator(NullLocator())

    def merge(self, other):
        """Simply discards the other guide"""
        return self

    def draw(self):
        """Nothing to do"""


class guide_prism(guide_prism_offset):
    """The prism offset guide that hides the minor ticks"""

    available_aes = {"x", "y"}

    def create_geoms(self, plot):
        """Apply the arguments"""
        gca = plot.axs[0].axes
        if self.aesthetic == "x":
            gca.xaxis.set_minor_locator(NullLocator())
        else:
            gca.yaxis.set_minor_locator(NullLocator())


class guide_prism_minor(guide_prism_offset):
    """The prism offset guide that shows the minor ticks"""

    available_aes = {"x", "y"}

    def create_geoms(self, plot):
        """Apply the arguments"""


class guide_prism_offset_minor(guide_prism_offset):
    """The prism offset guide that shows the minor ticks with offset"""

    def create_geoms(self, plot):
        """Apply the arguments"""
        xaxis_position = plot.axs[0].xaxis.get_label_position()
        yaxis_position = plot.axs[0].yaxis.get_label_position()
        gca = plot.axs[0].axes

        if self.aesthetic == "x":
            gca.spines[yaxis_position].set_position(("outward", self.offset))
            major_locs = gca.get_xticks()
            minor_locs = gca.xaxis.get_minorticklocs()
            min_loc = min(min(major_locs), min(minor_locs))
            max_loc = max(max(major_locs), max(minor_locs))
            gca.set_xlim(min_loc, max_loc)
        else:
            gca.spines[xaxis_position].set_position(("outward", self.offset))
            major_locs = gca.get_yticks()
            minor_locs = gca.yaxis.get_minorticklocs()
            min_loc = min(min(major_locs), min(minor_locs))
            max_loc = max(max(major_locs), max(minor_locs))
            gca.set_ylim(min_loc, max_loc)


# class AxesBracketDecorator:
#     def __init__(self, ax, ticks, size="5%", pad=0.05, spacing=0.05):

#         divider = make_axes_locatable(ax)
#         self.ax = divider.new_vertical(
#             size=size,
#             pad=pad,
#             sharex=ax,
#             pack_start=True,
#         )
#         ax.figure.add_axes(self.ax)
#         ax.xaxis.set_visible(False)
#         # ax.xaxis.spines['bottom'].set_visible(False)
#         # self.ax.tick_params(axis='x', which=u'both', length=0)
#         for direction in ["left", "right", "bottom", "top"]:
#             self.ax.spines[direction].set_visible(False)
#         self.ax.set_yticks([])
#         pos = self.ax.get_position()

#         self.dist = numpy.mean(numpy.diff(ticks))
#         self.spacing = spacing
#         # self.curve = self.get_curve()
#         for tick in ticks:
#             # self.plot_curve(tick)
#             pass


# class guide_prism_bracket(guide_prism_offset):

#     def create_geoms(self, plot):
#         gca = plot.axs[0].axes
#         if self.aesthetic == "x":
#             AxesBracketDecorator(gca, ticks=gca.get_xticks())
#         else:
#             AxesBracketDecorator(gca, ticks=gca.get_yticks())



"""Provides available palettes"""

from functools import lru_cache
from pathlib import Path

import rtoml
from diot import Diot

SCHEMES_DIR = Path(__file__).parent.joinpath("schemes")


@lru_cache()
def list_themes():
    """List all available theme palettes"""
    return [
        tfile.stem
        for tfile in SCHEMES_DIR.glob("*.toml")
        if not tfile.stem.startswith("_")
    ]


@lru_cache()
def _all_color_pals():
    with SCHEMES_DIR.joinpath("_color_palettes.toml").open() as fcolor:
        return rtoml.load(fcolor)


@lru_cache()
def _all_fill_pals():
    with SCHEMES_DIR.joinpath("_fill_palettes.toml").open() as ffill:
        return rtoml.load(ffill)


@lru_cache()
def _all_shape_pals():
    with SCHEMES_DIR.joinpath("_shape_palettes.toml").open() as fshape:
        return rtoml.load(fshape)


@lru_cache()
def theme_colors(palette):
    """Get the colors for a specific theme"""
    with SCHEMES_DIR.joinpath(f"{palette}.toml") as fsch:
        return Diot(rtoml.load(fsch))


def list_color_pals():
    """List all available color palettes"""
    return list(_all_color_pals())


def list_fill_pals():
    """List all available fill palettes"""
    return list(_all_fill_pals())


def list_shape_pals():
    """List all available shape palettes"""
    return list(_all_shape_pals())


def prism_color_pal(palette):
    """Get the prism color palette by name"""
    return lambda n: _all_color_pals()[palette][:n]


def prism_fill_pal(palette):
    """Get the prism fill palette by name"""
    return lambda n: _all_fill_pals()[palette][:n]


def prism_shape_pal(palette):
    """Get the prism shape palette by name"""
    return lambda n: _all_shape_pals()[palette][:n]


list_colour_pals = list_color_pals
prism_colour_pal = prism_color_pal
